# lsqfitgp/_jaxext/__init__.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import traceback
import functools

import jax
from jax import numpy as jnp

from ._batcher import batchufunc
from ._fasthash import fasthash64, fasthash32

def makejaxufunc(ufunc, *derivs, excluded=None, floatcast=False):
    """
    
    Wrap a numpy ufunc to add jax support.
    
    Parameters
    ----------
    ufunc : callable
        Elementwise function following numpy broadcasting and type promotion
        rules. Keyword arguments not supported.
    derivs : sequence of callable
        Derivatives of the function w.r.t. each positional argument, with the
        same signature as `ufunc`. Pass None to indicate a missing derivative.
        There must be as many derivatives as the arguments to `ufunc`.
    excluded : sequence of int, optional
        The indices of arguments that are not broadcasted.
    floatcast : bool, default False
        If True, cast all arguments to float before calling the ufunc.
    
    Return
    ------
    func : callable
        Wrapped `ufunc`. Supports jit, but the calculation is performed on cpu.
    
    """

    nondiff_argnums = [i for i, d in enumerate(derivs) if d is None]
    
    @functools.wraps(ufunc)
    @functools.partial(jax.custom_jvp, nondiff_argnums=nondiff_argnums)
    def func(*args):
        args = tuple(map(jnp.asarray, args))
        if floatcast:
            flt = float_type(*args)
            args = tuple(a.astype(flt) for a in args)
        return pure_callback_ufunc(ufunc, jnp.result_type(*args), *args, excluded=excluded)

    @func.defjvp
    def func_jvp(*allargs):
        ndargs = allargs[:-2]
        dargs = allargs[-2]
        dargst = allargs[-1]
        
        itnd = iter(ndargs)
        itd = iter(dargs)
        args = [next(itnd) if d is None else next(itd) for d in derivs]
        
        result = func(*args)
        tangent = sum([
            d(*args) * t for t, d in
            zip(dargst, (d for d in derivs if d is not None))
        ])
        return result, tangent
    
    return func

def elementwise_grad(fun, argnum=0):
    assert int(argnum) == argnum and argnum >= 0, argnum
    @functools.wraps(fun)
    def funderiv(*args, **kw):
        preargs = args[:argnum]
        postargs = args[argnum + 1:]
        def oneargfun(arg):
            args = preargs + (arg,) + postargs
            return fun(*args, **kw)
        primal = args[argnum]
        shape = getattr(primal, 'shape', ())
        dtype = getattr(primal, 'dtype', type(primal))
        tangent = jnp.ones(shape, dtype)
        primal_out, tangent_out = jax.jvp(oneargfun, (primal,), (tangent,))
        return tangent_out
    return funderiv

class skipifabstract:
    """
    Context manager to try to do all operations eagerly even during jit, and
    skip entirely if it is not possible.
    """
    # I feared this would be slow because of the slow jax exception handling,
    # but %timeit suggests it isn't

    ENSURE_COMPILE_TIME_EVAL = True
    ENABLED = True
    
    def __enter__(self):
        if self.ENSURE_COMPILE_TIME_EVAL and self.ENABLED:
            self.mgr = jax.ensure_compile_time_eval()
            self.mgr.__enter__()
    
    def __exit__(self, exc_type, exc_value, tb):
        if not self.ENABLED:
            return
        exit = None
        if self.ENSURE_COMPILE_TIME_EVAL:
            exit = self.mgr.__exit__(exc_type, exc_value, tb)
        ignorable_error = (
            exc_type is not None
            and issubclass(exc_type, (
                jax.errors.ConcretizationTypeError,
                jax.errors.TracerArrayConversionError,
                    # TODO why isn't this a subclass of the former like
                    # TracerBoolConversionError? Open an issue
            ))
        )
        if exit or ignorable_error:
            return True
        
        weird_cond = exc_type is IndexError and (
            traceback.extract_tb(tb)[-1].name in ('arg_info_pytree', '_origin_msg'),
        )
        if weird_cond: # pragma: no cover
            # TODO this ignores a jax internal bug I don't understand, appears
            # in examples/pdf4.py
            return True

def float_type(*args):
    t = jnp.result_type(*args)
    return jnp.sin(jnp.empty(0, t)).dtype
    # numpy does this with common_type, but that supports only arrays, not
    # dtypes in the input. jnp.common_type is not defined.

def is_jax_type(dtype):
    dtype = jnp.dtype(dtype)
    try:
        jnp.empty(0, dtype)
        return True
    except TypeError as e:
        if 'JAX only supports number and bool dtypes' in str(e):
            return False
        raise

def pure_callback_ufunc(callback, dtype, *args, excluded=None, **kwargs):
    """ version of jax.pure_callback that deals correctly with ufuncs,
    see https://github.com/google/jax/issues/17187 """
    if excluded is None:
        excluded = ()
    shape = jnp.broadcast_shapes(*(
        a.shape
        for i, a in enumerate(args)
        if i not in excluded
    ))
    ndim = len(shape)
    padded_args = [
        a if i in excluded
        else jnp.expand_dims(a, tuple(range(ndim - a.ndim)))
        for i, a in enumerate(args)
    ]
    result = jax.ShapeDtypeStruct(shape, dtype)
    return jax.pure_callback(callback, result, *padded_args, vectorized=True, **kwargs)

    # TODO when jax solves this, check version and piggyback on original if new

def limit_derivatives(x, n, error_func=None):
    """
    Limit the number of derivatives that goes through a value.

    Parameters
    ----------
    x : array_like
        The value.
    n : int
        The maximum number of derivatives allowed. Must be an integer.
    error_func : callable, optional
        A function that takes (derivatives taken, n) and returns an exception.

    Return
    ------
    x : array_like
        The value, unchanged.
    """
    assert n == int(n)
    if error_func is None:
        def error_func(current, n):
            return ValueError(f'took {current} derivatives > limit {n}')
    return _limit_derivatives_impl(0, n, error_func, x)

@functools.partial(jax.custom_jvp, nondiff_argnums=(0, 1, 2))
def _limit_derivatives_impl(current, limit, func, x):
    if current > limit:
        raise func(current, limit)
    return x

@_limit_derivatives_impl.defjvp
def _limit_derivatives_impl_jvp(current, limit, func, primals, tangents):
    x, = primals
    xdot, = tangents
    return _limit_derivatives_impl(current + 1, limit, func, x), xdot



# lsqfitgp/copula/_copula.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" defines Copula """

import functools
import pprint

import jax
from jax import tree_util
from jax import numpy as jnp
import gvar

from .. import _array
from .. import _gvarext
from . import _base

class Copula(_base.DistrBase):
    """
    
    Represents a tree of probability distributions.

    By "tree" it is intended an arbitrarily nested structure of containers
    (e.g., `dict` or `list`) where each leaf node is either a `Distr` object or
    another `Copula`.

    The function of a `Copula` is to keep into account the relationships
    between all the `Distr` objects when defining the function `partial_invfcn`
    that maps Normal variates to the desired random variable. The same `Distr`
    object appearing in multiple places will not be sampled more than once.

    This class inherits the functionality defined in `DistrBase` without
    additions. The attributes `shape`, `dtype`, `distrshape`, that represents
    properties of the output of `partial_invfcn`, are trees matching the
    structure of the tree of variables. The same holds for the output of
    `partial_invfcn`. `in_shape` instead is an ordinary array shape, indicating
    that the input Normal variables are a raveled single array.

    Parameters
    ----------
    variables : tree of Distr or Copula
        The tree of distributions to wrap. The containers are copied, so
        successive modifications will not be reflected on the `Copula`.

    See also
    --------
    DistrBase, Distr, jax.tree_util

    Examples
    --------

    Define a model into a dictionary one piece at a time, then wrap it as a
    `Copula`:

    >>> m = {}
    >>> m['a'] = lgp.copula.halfnorm(1.5)
    >>> m['b'] = lgp.copula.halfcauchy(2)
    >>> m['c'] = [
    ...     lgp.copula.uniform(m['a'], m['b']),
    ...     lgp.copula.uniform(m['b'], m['a']),
    ... ]
    >>> cop = lgp.copula.Copula(m)
    >>> cop
    Copula({'a': halfnorm(1.5),
     'b': halfcauchy(2),
     'c': [uniform(<a>, <b>), uniform(<b>, <a>)]})

    Notice how, when showing the object on the REPL, multiple appearances of
    the same variables are replaced by identifiers derived from the dictionary
    keys.

    The model may then be extended to create a variant. This does not affect
    `cop`:

    >>> m['d'] = lgp.copula.invgamma(m['c'][0], m['c'][1])
    >>> cop2 = lgp.copula.Copula(m)
    >>> cop2
    Copula({'a': halfnorm(1.5),
     'b': halfcauchy(2),
     'c': [uniform(<a>, <b>), uniform(<b>, <a>)],
     'd': invgamma(<c.0>, <c.1>)})
    >>> cop
    Copula({'a': halfnorm(1.5),
     'b': halfcauchy(2),
     'c': [uniform(<a>, <b>), uniform(<b>, <a>)]})

    """

    @staticmethod
    def _tree_path_str(path):
        """ format a jax pytree key path as a compact, readable string """
        def parsekey(key):
            if hasattr(key, 'key'):
                return key.key
            elif hasattr(key, 'idx'):
                return key.idx
            else:
                return key
        def keystr(key):
            key = parsekey(key)
            return str(key).replace('.', r'\.')
        return '.'.join(map(keystr, path))

    @classmethod
    def _jaxext_dict_sorting(cls, pytree):
        """ replace dicts in pytree with a custom dict subclass such their
        insertion order is maintained, see
        https://github.com/google/jax/issues/4085 """
        
        def is_dict(obj):
            return obj.__class__ is dict
        
        def patch_dict(obj):
            if is_dict(obj):
                return tree_util.tree_map(patch_dict, cls._Dict(obj))
            else:
                return obj
        
        return tree_util.tree_map(patch_dict, pytree, is_leaf=is_dict)

    @tree_util.register_pytree_with_keys_class
    class _Dict(dict):

        def tree_flatten_with_keys(self):
            treedef = dict.fromkeys(self)
            keys_values = [(tree_util.DictKey(k), v) for k, v in self.items()]
            return keys_values, treedef

        @classmethod
        def tree_unflatten(cls, treedef, values):
            return cls(zip(treedef, values))

    def __init__(self, variables):
        variables = self._jaxext_dict_sorting(variables)
        def check_type(path, obj):
            if not isinstance(obj, _base.DistrBase):
                raise TypeError(f'only Distr or Copula objects can be '
                    f'contained in a Copula, found {obj!r} at '
                    f'<{self._tree_path_str(path)}>')
            return obj
        self._variables = tree_util.tree_map_with_path(check_type, variables)
        cache = set()
        self.in_shape = self._compute_in_size(cache),
        self._ancestor_count = len(cache) - 1
        self.shape = self._map_getattr('shape')
        self.distrshape = self._map_getattr('distrshape')
        self.dtype = self._map_getattr('dtype')

    def _compute_in_size(self, cache):
        if (out := super()._compute_in_size(cache)) is not None:
            return out
        def accumulate(in_size, obj):
            return in_size + obj._compute_in_size(cache)
        return tree_util.tree_reduce(accumulate, self._variables, 0)

    def _map_getattr(self, attr):
        def get_attr(obj):
            if isinstance(obj, __class__):
                return obj._map_getattr(attr)
            else:
                return getattr(obj, attr)
        return tree_util.tree_map(get_attr, self._variables)

    def _partial_invfcn_internal(self, x, i, cache):
        if (out := super()._partial_invfcn_internal(x, i, cache)) is not None:
            return out
        
        distributions, treedef = tree_util.tree_flatten(self._variables)
        outputs = []
        for distr in distributions:
            out, i = distr._partial_invfcn_internal(x, i, cache)
            outputs.append(out)
        out = tree_util.tree_unflatten(treedef, outputs)

        cache[self] = out
        return out, i

    @functools.cached_property
    def _partial_invfcn(self):

        # non vectorized version, check core shapes and call recursive impl
        # @jax.jit
        def partial_invfcn_0(x):
            assert x.shape == self.in_shape
            cache = {}
            y, i = self._partial_invfcn_internal(x, 0, cache)
            assert i == x.size
            assert len(cache) == 1 + self._ancestor_count
            return y
        partial_invfcn_0_deriv = jax.jacfwd(partial_invfcn_0)

        # add 1-axis vectorization
        partial_invfcn_1 = jax.vmap(partial_invfcn_0)
        partial_invfcn_1_deriv = jax.vmap(partial_invfcn_0_deriv)

        # add gvar support
        def partial_invfcn_2(x):
            
            if x.dtype == object:
                
                # unpack the gvars
                in_mean = gvar.mean(x)
                in_jac, indices = _gvarext.jacobian(x)

                # apply function
                out_mean = partial_invfcn_1(in_mean)
                jac = partial_invfcn_1_deriv(in_mean)

                # concatenate derivatives and repack as gvars
                def contract_and_pack(out_mean, jac):
                    # indices:
                    # b = broadcast
                    # i = input
                    # ... = output
                    # g = gvar indices
                    out_jac = jnp.einsum('b...i,big->b...g', jac, in_jac)
                    return _gvarext.from_jacobian(out_mean, out_jac, indices)

                return tree_util.tree_map(contract_and_pack, out_mean, jac)

            else:
                return partial_invfcn_1(x)

        # add full vectorization
        def partial_invfcn_3(x):
            x = _array.asarray(x)
            assert x.shape[-1:] == self.in_shape
            head = x.shape[:-1]
            x = x.reshape((-1,) + self.in_shape)
            y = partial_invfcn_2(x)
            def reshape_y(y, shape):
                assert y.shape[1:] == shape
                y = y.reshape(head + shape)
                if y.dtype == object and not y.ndim:
                    y = y.item()
                return y
            return tree_util.tree_map(reshape_y, y, self.shape)

        return partial_invfcn_3

    def __repr__(self, path='', cache=None):
        
        if isinstance(cache := super().__repr__(path, cache), str):
            return cache
        
        def subrepr(k, obj):
            if isinstance(obj, _base.DistrBase):
                k = self._tree_path_str(k)
                return obj.__repr__('.'.join((path, k)).lstrip('.'), cache)
            else:
                return repr(obj)

        class NoQuotesRepr:
            def __init__(self, s):
                self.s = s
            def __repr__(self):
                return self.s
        
        out = tree_util.tree_map_with_path(subrepr, self._variables)
        out = tree_util.tree_map(NoQuotesRepr, out)
        out = pprint.pformat(out, sort_dicts=False)
        return f'{self.__class__.__name__}({out})'

    def _compute_staticdescr(self, path, cache):
        def compute(key, x):
            return x._compute_staticdescr(path + [key], cache)
        return tree_util.tree_map_with_path(compute, self._variables)



# lsqfitgp/_Kernel/_crosskernel.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import enum
import functools
import sys
import collections
import types
import abc
import warnings

import numpy
from jax import numpy as jnp

from .. import _array
from .. import _jaxext
from .. import _utils

from . import _util

@functools.lru_cache(maxsize=None)
def least_common_superclass(*classes):
    """
    Find a "least" common superclass. The class is searched in all the MROs,
    but the comparison is done with `issubclass` to support virtual inheritance.
    """
    mros = [c.__mro__ for c in classes]
    indices = [0] * len(mros)
    for i, mroi in enumerate(mros):
        for j, mroj in enumerate(mros):
            if i == j:
                continue
            while not issubclass(mroi[0], mroj[indices[j]]):
                indices[j] += 1
    idx = numpy.argmin(indices)
    return mros[idx][indices[idx]]

class CrossKernel:
    r"""
    
    Base class to represent kernels, i.e., covariance functions.

    A kernel is a two-argument function that computes the covariance between
    two functions at some points according to some probability distribution:

    .. math::
        \mathrm{kernel}(x, y) = \mathrm{Cov}[f(x), g(y)].
    
    `CrossKernel` objects are callable, the signature is ``obj(x, y)``, and
    they can be summed and multiplied between them and with scalars. They
    are immutable; all operations return new objects.

    Parameters
    ----------
    core : callable
        A function with signature ``core(x, y)``, where ``x`` and ``y``
        are two broadcastable numpy arrays, which computes the value of the
        kernel.
    scale, loc, derivable, maxdim, dim :
        If specified, these arguments are passed as arguments to the
        correspondingly named operators, in the order listed here. See `linop`.
        Briefly: the kernel selects only fields `dim` in the input, checks the
        dimensionality against `maxdim`, checks there are not too many
        derivatives taken on the arguments, then transforms as ``(x - loc) /
        scale``. If any argument is callable, it is passed `**kw` and must
        return the actual argument. If an argument is a tuple, it is interpreted
        as a pair of arguments.
    forcekron : bool, default False
        If True, apply ``.transf('forcekron')`` to the kernel, before the
        operations above. Available only for `Kernel`.
    batchbytes : number, optional
        If specified, apply ``.batch(batchbytes)`` to the kernel.
    dynkw : dict, optional
        Additional keyword arguments passed to `core` that can be modified
        by transformations. Deleted by transformations by default.
    **initkw :
        Additional keyword arguments passed to `core` that can be read but not
        changed by transformations.

    Attributes
    ----------
    initkw : dict
        The `initkw` argument.
    dynkw : dict
        The `dynkw` argument, or a modification of it if the object has been
        transformed.
    core : callable
        The `core` argument partially evaluated on `initkw`, or another
        function wrapping it if the object has been transformed.
    
    Methods
    -------
    batch
    linop
    algop
    transf
    register_transf
    register_linop
    register_corelinop
    register_xtransf
    register_algop
    register_ufuncalgop
    transf_help
    has_transf
    list_transf
    super_transf
    make_linop_family

    See also
    --------
    Kernel

    Notes
    -----
    The predefined class hierarchy and the class logic of the transformations
    assume that each kernel class corresponds to a subalgebra, i.e., addition
    and multiplication preserve the class.
    
    """

    __slots__ = '_initkw', '_dynkw', '_core'
        # only __new__ and _clone shall access these attributes

    @property
    def initkw(self):
        return types.MappingProxyType(self._initkw)

    @property
    def dynkw(self):
        return types.MappingProxyType(self._dynkw)

    @property
    def core(self):
        return self._core

    def __new__(cls, core, *,
        scale=None,
        loc=None,
        derivable=None,
        maxdim=None,
        dim=None,
        forcekron=False,
        batchbytes=None,
        dynkw={},
        **initkw,
    ):
        self = super().__new__(cls)
                
        self._initkw = initkw
        self._dynkw = dict(dynkw)
        self._core = lambda x, y, **dynkw: core(x, y, **initkw, **dynkw)

        if forcekron:
            self = self.transf('forcekron')

        linop_args = {
            'scale': scale,
            'loc': loc,
            'derivable': derivable,
            'maxdim': maxdim,
            'dim': dim,
        }
        for transfname, arg in linop_args.items():
            if callable(arg):
                arg = arg(**initkw)
            if isinstance(arg, tuple):
                self = self.linop(transfname, *arg)
            else:
                self = self.linop(transfname, arg)

        if batchbytes is not None:
            self = self.batch(batchbytes)

        return self

    def __call__(self, x, y):
        x = _array.asarray(x)
        y = _array.asarray(y)
        shape = _array.broadcast(x, y).shape
        result = self.core(x, y, **self.dynkw)
        assert isinstance(result, (numpy.ndarray, jnp.number, jnp.ndarray))
        assert jnp.issubdtype(result.dtype, jnp.number), result.dtype
        assert result.shape == shape, (result.shape, shape)
        return result

    def _clone(self, cls=None, *, initkw=None, dynkw=None, core=None):
        newself = object.__new__(self.__class__ if cls is None else cls)
        newself._initkw = self._initkw if initkw is None else dict(initkw)
        newself._dynkw = {} if dynkw is None else dict(dynkw)
        newself._core = self._core if core is None else core
        return newself

    class _side(enum.Enum):
        LEFT = 0
        RIGHT = 1
         
    @classmethod
    def _nary(cls, op, kernels, side):
        
        if side is cls._side.LEFT:
            wrapper = lambda c, _, y, **kw: lambda x: c(x, y, **kw)
            arg = lambda x, _: x
        elif side is cls._side.RIGHT:
            wrapper = lambda c, x, _, **kw: lambda y: c(x, y, **kw)
            arg = lambda _, y: y
        else: # pragma: no cover
            raise KeyError(side)
        
        cores = [k.core for k in kernels]
        def core(x, y, **kw):
            wrapped = [wrapper(c, x, y, **kw) for c in cores]
            transformed = op(*wrapped)
            return transformed(arg(x, y))
        
        return __class__(core)

        # TODO instead of wrapping the cores just by closing over an argument,
        # define a `PartialKernel` object that has linop defined, with one arg
        # fixed to None, and __call__ that closes over an argument. This way I
        # could apply ops directly with deflintransf.

    # TODO move in the binary op methods the logic that decides wether to
    # return NotImplemented, while the algops will raise an exception if the
    # argument is not to their liking

    def __add__(self, other):
        return self.algop('add', other)

    __radd__ = __add__

    def __mul__(self, other):
        return self.algop('mul', other)

    __rmul__ = __mul__

    def __pow__(self, other):
        return self.algop('pow', exponent=other)

    def __rpow__(self, other):
        return self.algop('rpow', base=other)
    
    def _swap(self):
        """ permute the arguments """
        core = self.core
        return self._clone(
            __class__,
            core=lambda x, y, **kw: core(y, x, **kw),
        )

        # TODO make _swap a transf inherited by CrossIsotropicKernel => messes
        # up with Kernel subclasses. I want to make Kernel subclasses identity
        # on swap, so it can't be an inherited transf, because Kernel appears
        # as the second base.

    def batch(self, maxnbytes):
        """
        Return a batched version of the kernel.

        The batched kernel processes its inputs in chunks to try to limit memory
        usage.

        Parameters
        ----------
        maxnbytes : number
            The maximum number of input bytes per chunk, counted after
            broadcasting the input shapes. Actual broadcasting may not occur if
            not induced by the operations in the kernel.

        Returns
        -------
        batched_kernel : CrossKernel
            The same kernel but with batched computations.
        """
        core = _jaxext.batchufunc(self.core, maxnbytes=maxnbytes)
        return self._clone(core=core)
    
    @classmethod
    def _crossmro(cls):
        """ MRO iterator excluding subclasses of Kernel """
        for c in cls.mro(): # pragma: no branch
            if not issubclass(c, Kernel):
                yield c
            if c is __class__:
                break

    _transf = {}

    _Transf = collections.namedtuple('_Transf', ['func', 'doc', 'kind'])

    def __init_subclass__(cls, **kw):
        super().__init_subclass__(**kw)
        cls._transf = {}
        cls.__slots__ = ()

    @classmethod
    def _transfmro(cls):
        """ Iterator of superclasses with a _transf attribute """
        for c in cls.mro(): # pragma: no branch
            yield c
            if c is __class__:
                break

    @classmethod
    def _settransf(cls, transfname, transf):
        if transfname in cls._transf:
            raise KeyError(f'transformation {transfname!r} already registered '
                f'for {cls.__name__}')
        cls._transf[transfname] = cls._Transf(*transf)

    @classmethod
    def _alltransf(cls):
        """ list all accessible transfs as dict name -> (tcls, transf) """
        transfs = {}
        for tcls in cls._transfmro():
            for name, transf in tcls._transf.items():
                transfs.setdefault(name, (tcls, transf))
        return transfs

    @classmethod
    def _gettransf(cls, transfname, transfmro=None):
        """
        Find a transformation.

        The transformation is searched following the MRO up to `CrossKernel`.

        Parameters
        ----------
        transfname : hashable
            The transformation name.

        Returns
        -------
        cls : type
            The class where the transformation was found.
        transf, doc, kind : tuple
            The objects set by `register_transf`.

        Raises
        ------
        KeyError :
            The transformation was not found.
        """
        if transfmro is None:
            transfmro = cls._transfmro()
        for c in transfmro:
            try:
                return c, c._transf[transfname]
            except KeyError:
                pass
        raise KeyError(transfname)

    @classmethod
    def inherit_transf(cls, transfname, *, intermediates=False):
        """
        
        Inherit a transformation from a superclass.

        Parameters
        ----------
        transfname : hashable
            The name of the transformation.
        intermediates : bool, default False
            If True, make all superclasses up to the one definining the
            transformation inherit it too.

        Raises
        ------
        KeyError :
            The transformation was not found in any superclass, or the
            transformation is already registered on any of the target classes.

        See also
        --------
        transf

        """
        tcls, transf = cls._gettransf(transfname)
        cls._settransf(transfname, transf)
        if intermediates:
            for c in cls.mro()[1:]: # pragma: no branch
                if c is tcls:
                    break
                c._settransf(transfname, transf)

    @classmethod
    def inherit_all_algops(cls, intermediates=False):
        """

        Inherit all algebraic operations from superclasses.

        This makes sense if the class represents a subalgebra, i.e., it
        should be preserved by addition and multiplication.

        Parameters
        ----------
        intermediates : bool, default False
            If True, make all superclasses up to the one definining the
            transformation inherit it too.

        Raises
        ------
        KeyError :
            An algebraic operation is already registered for one of the target
            classes.

        See also
        --------
        transf

        """
        mro = cls._transfmro()
        next(mro)
        for name, (_, transf) in next(mro)._alltransf().items():
            if transf.kind is cls._algopmarker:
                cls.inherit_transf(name, intermediates=intermediates)

    @classmethod
    def list_transf(cls, superclasses=True):
        """
        List all the available transformations.

        Parameters
        ----------
        superclasses : bool, default True
            Include transformations defined in superclasses.

        Returns
        -------
        transfs: dict of Transf
            The dictionary keys are the transformation names, the values are
            named tuples ``(tcls, kind, impl, doc)`` where ``tcls`` is the class
            defining the transformation, ``kind`` is the kind of transformation,
            ``impl`` is the implementation with signature ``impl(tcls, self,
            *args, **kw)``, ``doc`` is the docstring.
        """
        if superclasses:
            source = cls._alltransf().items
        else:
            def source():
                for name, transf in cls._transf.items():
                    yield name, (cls, transf)
        return {
            name: cls.Transf(tcls, transf.kind, transf.func, transf.doc)
            for name, (tcls, transf) in source()
        }

    Transf = collections.namedtuple('Transf', ['tcls', 'kind', 'func', 'doc'])

    @classmethod
    def has_transf(cls, transfname):
        """
        Check if a transformation is registered.

        Parameters
        ----------
        transfname : hashable
            The transformation name.

        Returns
        -------
        has_transf : bool
            Whether the transformation is registered.

        See also
        --------
        transf
        """
        try:
            cls._gettransf(transfname)
        except KeyError as exc:
            if exc.args == (transfname,):
                return False
            else: # pragma: no cover
                raise
        else:
            return True

    @classmethod
    def transf_help(cls, transfname):
        """
        
        Return the documentation of a transformation.

        Parameters
        ----------
        transfname : hashable
            The name of the transformation.

        Returns
        -------
        doc : str
            The documentation of the transformation.

        See also
        --------
        transf

        """
        _, transf = cls._gettransf(transfname)
        return transf.doc

    def transf(self, transfname, *args, **kw):
        """

        Return a transformed kernel.

        Parameters
        ----------
        transfname : hashable
            A name identifying the transformation.
        *args, **kw :
            Arguments to the transformation.

        Returns
        -------
        newkernel : object
            The output of the transformation.

        Raises
        ------
        KeyError
            The transformation is not defined in this class or any superclass.

        See also
        --------
        linop, algop, transf_help, has_transf, list_transf, super_transf, register_transf, register_linop, register_corelinop, register_xtransf, register_algop, register_ufuncalgop

        """
        tcls, transf = self._gettransf(transfname)
        return transf.func(tcls, self, *args, **kw)

    @classmethod
    def super_transf(cls, transfname, self, *args, **kw):
        """
        
        Transform the kernel using a superclass transformation.

        This is equivalent to `transf` but is invoked on a class and the
        object is passed explicitly. The definition of the transformation is
        searched starting from the next class in the MRO of `self`.

        Parameters
        ----------
        transfname, *args, **kw :
            See `transf`.
        self : CrossKernel
            The object to transform.

        Returns
        -------
        newkernel : object
            The output of the transformation.

        """
        mro = list(self._transfmro())
        idx = mro.index(cls)
        tcls, transf = self._gettransf(transfname, mro[idx + 1:])
        return transf.func(tcls, self, *args, **kw)

    def linop(self, transfname, *args, **kw):
        r"""

        Transform kernels to represent the application of a linear operator.

        .. math::
            \text{kernel}_1(x, y) &= \mathrm{Cov}[f_1(x), g_1(y)], \\
            \text{kernel}_2(x, y) &= \mathrm{Cov}[f_2(x), g_2(y)], \\
            &\ldots \\
            \text{newkernel}(x, y) &=
                \mathrm{Cov}[T_f(f_1, f_2, \ldots)(x), T_g(g_1, g_2, \ldots)(y)]

        Parameters
        ----------
        transfname : hashable
            The name of the transformation.
        *args :
            A sequence of `CrossKernel` instances, representing the operands,
            followed by one or two non-kernel arguments, indicating how to act
            on each side of the kernels. If both arguments represent the
            identity, this is a no-op. If there is only one argument, it is
            intended that the two arguments are equal. `None` always represents
            the identity.

        Returns
        -------
        newkernel : CrossKernel
            The transformed kernel.

        Raises
        ------
        ValueError :
            The transformation exists but was not defined by `register_linop`.

        See also
        --------
        transf

        Notes
        -----
        The linear operator is defined on the function the kernel represents the
        distribution of, not in full generality on the kernel itself. When
        multiple kernels are involved, their distributions may be considered
        independent or not, depending on the specific operation.

        If the result is a subclass of the class defining the transformation,
        the result is casted to the latter. Then, if the result and all the
        operands are instances of `Kernel`, but the two operator arguments
        differ, the result is casted to its first non-`Kernel` superclass.
        
        """
        tcls, transf = self._gettransf(transfname)
        if transf.kind is not self._linopmarker:
            raise ValueError(f'the transformation {transfname!r} was not '
                f'defined with register_linop and so can not be invoked '
                f'by linop')
        return transf.func(tcls, self, *args)

    def algop(self, transfname, *operands, **kw):
        r"""

        Return a nonnegative algebraic transformation of the input kernels.

        .. math::
            \mathrm{newkernel}(x, y) &=
                f(\mathrm{kernel}_1(x, y), \mathrm{kernel}_2(x, y), \ldots), \\
            f(z_1, z_2, \ldots) &= \sum_{k_1,k_2,\ldots=0}^\infty
                a_{k_1 k_2 \ldots} z_1^{k_1} z_2^{k_2} \ldots,
            \quad a_* \ge 0.

        Parameters
        ----------
        transfname : hashable
            A name identifying the transformation.
        *operands : CrossKernel, scalar
            Arguments to the transformation in addition to self.
        **kw :
            Additional arguments to the transformation, not considered as
            operands.

        Returns
        -------
        newkernel : CrossKernel or NotImplemented
            The transformed kernel, or NotImplemented if the operation is
            not supported.

        See also
        --------
        transf

        Notes
        -----
        The class of `newkernel` is the least common superclass of: the
        "natural" output of the operation, the class defining the
        transformation, and the classes of the operands.

        For class determination, scalars in the input count as `Constant`
        if nonnegative or traced by jax, else `CrossConstant`.

        """
        tcls, transf = self._gettransf(transfname)
        if transf.kind is not self._algopmarker:
            raise ValueError(f'the transformation {transfname!r} was not '
                f'defined with register_algop and so can not be invoked '
                f'by algop')
        return transf.func(tcls, self, *operands, **kw)

    @classmethod
    def register_transf(cls, func, transfname=None, doc=None, kind=None):
        """
        
        Register a transformation for use with `transf`.

        The transformation will be accessible to subclasses.

        Parameters
        ----------
        func : callable
            A function ``func(tcls, self, *args, **kw) -> object`` implementing
            the transformation, where ``tcls`` is the class that defines the
            transformation.
        transfname : hashable, optional.
            The `transfname` parameter to `transf` this transformation will
            be accessible under. If not specified, use the name of `func`.
        doc : str, optional
            The documentation of the transformation for `transf_help`. If not
            specified, use the docstring of `func`.
        kind : object, optional
            An arbitrary marker.

        Returns
        -------
        func : callable
            The `func` argument as is.

        Raises
        ------
        KeyError :
            The name is already in use for another transformation in the same
            class.

        See also
        --------
        transf

        """
        if transfname is None:
            transfname = func.__name__
        if doc is None:
            doc = func.__doc__
        cls._settransf(transfname, (func, doc, kind))
        return func

        # TODO forbid to override with a different kind?

    @classmethod
    def register_linop(cls, op, transfname=None, doc=None, argparser=None):
        """
        
        Register a transformation for use with `linop`.

        Parameters
        ----------
        op : callable
            A function ``op(tcls, self, arg1, arg2, *operands) -> CrossKernel``
            that returns the new kernel, where ``arg1`` and ``arg2`` represent
            the operators acting on each side of the kernels, and ``operands``
            are the other kernels beyond ``self``.
        transfname, doc : optional
            See `register_transf`.
        argparser : callable, optional
            A function applied to ``arg1`` and ``arg2``. Not called if the
            argument is `None`. It should map the identity to `None`.

        Returns
        -------
        op : callable
            The `op` argument as is.

        Notes
        -----
        The function `op` is called only if ``arg1`` or ``arg2`` is not `None`
        after potential conversion with `argparser`.

        See also
        --------
        transf

        """
        
        if transfname is None:
            transfname = op.__name__ # for result type error message
        
        @functools.wraps(op)
        def func(tcls, self, *allargs):

            # split the arguments in kernels and non-kernels
            for pos, arg in enumerate(allargs):
                if not isinstance(arg, __class__):
                    break
            else:
                pos = len(allargs)
            operands = allargs[:pos]
            args = allargs[pos:]
            
            # check the arguments from the first non-kernel onwards are 1 or 2
            if len(args) not in (1, 2):
                raise ValueError(f'incorrect number of non-kernel tail '
                    f'arguments {len(args)}, expected 1 or 2')

            # wrap argument parser to enforce preserving None
            if argparser:
                conv = lambda x: None if x is None else argparser(x)
            else:
                conv = lambda x: x
            
            # determine if the two arguments count as "identical" or not
            if len(args) == 1:
                arg = conv(*args)
                different = False
                arg1 = arg2 = arg
            else:
                arg1, arg2 = args
                different = arg1 is not arg2
                arg1 = conv(arg1)
                arg2 = conv(arg2)
                different &= arg1 is not arg2
                    # they must be not identical both before and after to handle
                    # these cases:
                    #  - if the user passes identical arguments, but argparser
                    #    makes copies, it must still count as identical
                    #  - if the user passes arguments which are not identical,
                    #    but argparser sends them to the same object, then they
                    #    surely represent the same transf

            # handle no-op case
            if arg1 is None and arg2 is None:
                return self

            # invoke implementation
            result = op(tcls, self, arg1, arg2, *operands)

            # check result is a kernel
            if not isinstance(result, __class__):
                raise TypeError(f'linop {transfname!r} returned '
                    f'object of type {result.__class__.__name__}, expected '
                    f'subclass of {__class__.__name__}')
            
            # modify class of the result
            rcls = result.__class__
            if issubclass(rcls, tcls):
                rcls = tcls
            all_operands_kernel = all(isinstance(o, Kernel) for o in operands)
            if isinstance(self, Kernel) and all_operands_kernel and different:
                rcls = next(rcls._crossmro())
            if rcls is not result.__class__:
                result = result._clone(rcls)
            
            return result

        cls.register_transf(func, transfname, doc, cls._linopmarker)
        return op

    class _LinOpMarker(str): pass
    _linopmarker = _LinOpMarker('linop')

    @classmethod
    def register_corelinop(cls, corefunc, transfname=None, doc=None, argparser=None):
        """

        Register a linear operator with a function that acts only on the core.

        Parameters
        ----------
        corefunc : callable
            A function ``corefunc(core, arg1, arg2, *cores) -> newcore``, where
            ``core`` is the function that implements the kernel passed at
            initialization, and ``cores`` for other operands.
        transfname, doc, argparser :
            See `register_linop`.

        Returns
        -------
        corefunc : callable
            The `corefunc` argument as is.

        See also
        --------
        transf

        """
        @functools.wraps(corefunc)
        def op(_, self, arg1, arg2, *operands):
            cores = (o.core for o in operands)
            core = corefunc(self.core, arg1, arg2, *cores)
            return self._clone(core=core)
        cls.register_linop(op, transfname, doc, argparser)
        return corefunc

    @classmethod
    def register_xtransf(cls, xfunc, transfname=None, doc=None):
        """

        Register a linear operator that acts only on the input.

        Parameters
        ----------
        xfunc : callable
            A function ``xfunc(arg) -> (lambda x: newx)`` that takes in a
            `linop` argument and produces a function to transform the input. Not
            called if ``arg`` is `None`. To indicate the identity, return
            `None`.
        transfname, doc :
            See `register_linop`. `argparser` is not provided because its
            functionality can be included in `xfunc`.

        Returns
        -------
        xfunc : callable
            The `xfunc` argument as is.

        See also
        --------
        transf

        """

        @functools.wraps(xfunc)
        def corefunc(core, xfun, yfun):
            if not xfun:
                return lambda x, y, **kw: core(x, yfun(y), **kw)
            elif not yfun:
                return lambda x, y, **kw: core(xfun(x), y, **kw)
            else:
                return lambda x, y, **kw: core(xfun(x), yfun(y), **kw)
        
        cls.register_corelinop(corefunc, transfname, doc, xfunc)
        return xfunc

    @classmethod
    def register_algop(cls, op, transfname=None, doc=None):
        """

        Register a transformation for use with `algop`.

        Parameters
        ----------
        op : callable
            A function ``op(tcls, *kernels, **kw) -> CrossKernel |
            NotImplemented`` that returns the new kernel. ``kernels`` may be
            scalars but for the first argument.
        transfname, doc :
            See `register_transf`.

        Returns
        -------
        op : callable
            The `op` argument as is.

        See also
        --------
        transf

        """

        if transfname is None:
            transfname = op.__name__ # for error message
        
        @functools.wraps(op)
        def func(tcls, *operands, **kw):
            result = op(tcls, *operands, **kw)
            
            if result is NotImplemented:
                return result
            elif not isinstance(result, __class__):
                raise TypeError(f'algop {transfname!r} returned '
                    f'object of type {result.__class__.__name__}, expected '
                    f'subclass of {__class__.__name__}')
            
            def classes():
                yield tcls
                for o in operands:
                    if isinstance(o, __class__):
                        yield o.__class__
                    elif _util.is_nonnegative_scalar_trueontracer(o):
                        yield Constant
                    elif _util.is_numerical_scalar(o):
                        yield CrossConstant
                    else:
                        raise TypeError(f'operands to algop {transfname!r} '
                            f'must be CrossKernel or numbers, found {o!r}')
                        # this type check comes after letting the implementation
                        # return NotImplemented, to support overloading
                yield result.__class__
            
            lcs = least_common_superclass(*classes())
            return result._clone(lcs)
    
        cls.register_transf(func, transfname, doc, cls._algopmarker)
        return op

        # TODO delete initkw (also in linop) if there's more than one kernel
        # operand or if the class changed?

        # TODO consider adding an option domains=callable, returns list of
        # domains from the operands, or list of tuples right away, and the
        # impl checks at runtime (if not traced) that the output values are in
        # the domains, with an informative error message

        # TODO consider escalating to ancestor definitions of the transf
        # until an impl does not return NotImplemented, and then raise an
        # an error instead of letting the NotImplemented escape. This would be
        # useful to partial behavior ovverride in subclasses, e.g, handle
        # multiplication by scalar to rewire transformations.

    class _AlgOpMarker(str): pass
    _algopmarker = _AlgOpMarker('algop')

    @classmethod
    def register_ufuncalgop(cls, ufunc, transfname=None, doc=None):
        """

        Register an algebraic operation with a function that acts only on the
        kernel value.

        Parameters
        ----------
        corefunc : callable
            A function ``ufunc(*values, **kw) -> value``, where ``values`` are
            the values yielded by the operands.
        transfname, doc :
            See `register_transf`.

        Returns
        -------
        ufunc : callable
            The `ufunc` argument as is.

        See also
        --------
        transf

        """
        @functools.wraps(ufunc)
        def op(_, self, *operands, **kw):
            cores = tuple(
                o.core if isinstance(o, __class__)
                else lambda x, y: o
                for o in (self, *operands)
            )
            def core(x, y, **kw):
                values = (core(x, y, **kw) for core in cores)
                return ufunc(*values, **kw)
            return self._clone(core=core)
        cls.register_algop(op, transfname, doc)
        return ufunc

    @classmethod
    def make_linop_family(cls, transfname, bothker, leftker, rightker=None, *,
        doc=None, argparser=None, argnames=None, translkw=None):
        """
        
        Form a family of kernels classes related by linear operators.

        The class this method is called on is the seed class. A new
        trasformation is registered to obtain other classes from the seed class
        by applying a newly defined linop transformation.

        Parameters
        ----------
        transfname : str
            The name of the new transformation.
        bothker, leftker, rightker : CrossKernel
            The kernel classes to be obtained by applying the operator to a seed
            class object respectively on both sides, only left, or only right.
            All classes are assumed to require no positional arguments at
            construction, and recognize the same set of keyword arguments. If
            `rightker` is not specified, it is defined by subclassing `leftker`
            and transposing the kernel on object construction.
        doc, argparser : callable, optional
            See `register_linop`.
        argnames : pair of str, optional
            If specified, `leftker` is passed an additional keyword argument
            with name ``argnames[0]`` that specifies the argument of the
            operator, `rightker` is passed ``argnames[1]``, and `bothker`
            similarly is passed both.

            If `leftker` is used to implement `rightker`, it must accept both
            arguments, and may use them to determine if it is being invoked as
            `leftker` or `rightker`.
        translkw : callable, optional
            A function with signature ``translkw(dynkw, <argnames>, **initkw) ->
            dict`` that determines the constructor arguments for a new object
            starting from the ones of the object to be transformed. By default,
            ``initkw`` is passed over, and an error is raised if ``dynkw`` is
            not empty.

        Examples
        --------

        >>> @lgp.kernel
        ... def A(x, y, *, gatto):
        ...     ''' The reknown A kernel of order gatto '''
        ...     return ...
        ...
        >>> @lgp.kernel
        ... def T(n, m, *, gatto, op):
        ...     ''' The kernel of the Topo series of an A process of order
        ...         gatto '''
        ...     op1, op2 = op
        ...     return ...
        ...
        >>> @lgp.crosskernel
        ... def CrossTA(n, x, *, gatto, op):
        ...     ''' The cross covariance between the Topo series of an A
        ...         process of order gatto and the process itself '''
        ...     return ...
        ...
        >>> A.make_linop_family('topo', T, CrossTA)
        >>> a = A(gatto=7)
        >>> ta = A.linop('topo', True, None)
        >>> at = A.linop('topo', None, True)
        >>> t = A.linop('topo', True)

        See also
        --------
        transf

        """
        
        if rightker is None:

            # invent a name for rightker
            rightname = f'Cross{cls.__name__}{bothker.__name__}'

            # define how to set up rightker
            def exec_body(ns):
                
                if leftker.__doc__:
                    header = 'Automatically generated transposed version of:\n\n'
                    ns['__doc__'] = _utils.append_to_docstring(leftker.__doc__, header, front=True)
                
                def __new__(cls, *args, **kw):
                    self = super(rightker, cls).__new__(cls, *args, **kw)
                    
                    if self.__class__ is cls:
                        self = self._swap()
                        if not isinstance(self, leftker):
                            raise TypeError(f'newly created instance of '
                                f'automatically defined {rightker.__name__} is not an '
                                f'instance of {leftker.__name__} after '
                                f'transposition. Either define transposition '
                                f'for {leftker.__name__}, or define '
                                f'{rightker.__name__} manually')
                        return self._clone(cls)

                    else:
                        return self._swap()

                ns['__new__'] = __new__

            # create rightker, evil twin of leftker separated at birth
            rightker = types.new_class(rightname, (leftker,), exec_body=exec_body)

        # check which classes are symmetric
        classes = cls, bothker, leftker, rightker
        sym = tuple(issubclass(c, Kernel) for c in classes)
        exp = True, True, False, False
        if sym != exp:
            desc = lambda t: 'Kernel' if t else 'non-Kernel'
            warnings.warn(f'Expected classes pattern {", ".join(map(desc, exp))}, '
                f'found {", ".join(map(desc, sym))}')

        # set translkw if not specified
        if translkw is None:
            def translkw(*, dynkw, **initkw):
                if dynkw:
                    raise ValueError('found non-empty `dynkw`, the default '
                        'implementation of `translkw` does not support it')
                return initkw

        # function to produce the arguments to the transformed objects
        def makekw(self, arg1, arg2):
            kw = dict(dynkw=self.dynkw, **self.initkw)
            if argnames is not None:
                if arg1 is not None:
                    kw = dict(**kw, **{argnames[0]: arg1})
                if arg2 is not None:
                    kw = dict(**kw, **{argnames[1]: arg2})
            return translkw(**kw)

        # register linop mapping cls to either leftker, rightker or bothker
        regkw = dict(transfname=transfname, doc=doc, argparser=argparser)
        @functools.partial(cls.register_linop, **regkw)
        def op_seed_to_siblings(_, self, arg1, arg2):
            kw = makekw(self, arg1, arg2)
            if arg2 is None:
                return leftker(**kw)
            elif arg1 is None:
                return rightker(**kw)
            else:
                return bothker(**kw)

        # register linop mapping leftker to bothker
        @functools.partial(leftker.register_linop, **regkw)
        def op_left_to_both(_, self, arg1, arg2):
            if arg1 is None:
                return bothker(**makekw(self, arg1, arg2))
            else:
                raise ValueError(f'cannot further transform '
                    f'`{leftker.__name__}` on left side with linop '
                    f'{transfname!r}')

        # register linop mapping rightker to bothker
        @functools.partial(rightker.register_linop, **regkw)
        def op_right_to_both(_, self, arg1, arg2):
            if arg2 is None:
                return bothker(**makekw(self, arg1, arg2))
            else:
                raise ValueError(f'cannot further transform '
                    f'`{rightker.__name__}` on right side with linop '
                    f'{transfname!r}')

class AffineSpan(CrossKernel, abc.ABC):
    """

    Kernel that tracks affine transformations.

    An `AffineSpan` instance accumulates the overall affine transformation
    applied to its inputs and output.

    `AffineSpan` and it subclasses are preserved by the transformations
    'scale', 'loc', 'add' (with scalar) and 'mul' (with scalar).

    `AffineSpan` can not be instantiated directly or used as standalone
    superclass. It must be the first base before concrete superclasses.

    """
    
    _affine_dynkw = dict(lloc=0, rloc=0, lscale=1, rscale=1, offset=0, ampl=1)

    def __new__(cls, *args, dynkw={}, **kw):
        if cls is __class__:
            raise TypeError(f'cannot instantiate {__class__.__name__} directly')
        new_dynkw = dict(cls._affine_dynkw)
        new_dynkw.update(dynkw)
        return super().__new__(cls, *args, dynkw=new_dynkw, **kw)

    def __init_subclass__(cls, **kw):
        super().__init_subclass__(**kw)
        for name in __class__._transf:
            cls.inherit_transf(name)

        # TODO I would like to inherit conditional on there not being another
        # definition. However, since registrations are done after class
        # creation, this method is invoked too early to do that.
        #
        # Is it possible to reimplement the transformation system to have the
        # transformations defined in the class body?
        #
        # Option 1: use a decorator that marks the transf methods and makes
        # them staticmethods. Then an __init_subclass__ above CrossKernel goes
        # through the methods and registers the marked ones.
        # 
        # Option 2: try to use descriptors (classes like property). Is it
        # possible to reimplement from scratch something like classmethod? If
        # so, I could make a transfmethod that bounds two arguments: tcls
        # for the class that defines it, and self for the object that invokes
        # it. Then inheriting would be just copying the thing over. If I wanted
        # to keep the invocation through method interface, I could define the
        # methods with underscores and register them somewhere.
        #
        # However, since I define kernels with decorators, I can't actually
        # count on the transformations being defined into them at class
        # creation. But that would be quite a corner case.
        #
        # To make subclassing convenient without decorators, allow to define a
        # core(x, y) method, which __init_subclass__ converts to staticmethod,
        # and have __new__ use it to set _core (I can't use the same name with
        # slots).
        #
        # to use without post-class registering make_linopfamily, make it a
        # decorator with attributed subdecorators for other members:
        #
        # @linop_family('T')
        # @kernel
        # def A(...)
        # @A.left('arg')
        # @crosskernel
        # def CrossTA(...)
        # @A.right # optional, generated from left if not specified
        # ...
        # @A.both('arg1', 'arg2')
        # @kernel
        # def T(...)
        #
        # Can also be stacked on top of left/right/both to chain families. Works
        # both with decorators and explicit class definitions.

    def _clone(self, *args, **kw):
        newself = super()._clone(*args, **kw)
        if isinstance(newself, __class__):
            for name in self._affine_dynkw:
                newself._dynkw[name] = self._dynkw[name]
        return newself

    @classmethod
    def __subclasshook__(cls, sub):
        if cls is __class__:
            return NotImplemented
                # to avoid algops promoting to unqualified AffineSpan
        if issubclass(cls, Kernel):
            if issubclass(sub, Constant):
                return True
            else:
                return NotImplemented
        elif issubclass(sub, CrossConstant):
            return True
        else:
            return NotImplemented

    # TODO I could do separately AffineLeft, AffineRight, AffineOut, and make
    # this a subclass of those three. AffineOut would also allow keeping the
    # class when adding two objects without keyword arguments in initkw and
    # dynkw beyond those managed by Affine. => Make only AffineSide and have it
    # switch side on _swap.

    # TODO when I reimplement transformations as methods, make AffineSpan not
    # a subclass of CrossKernel. Right now I have to to avoid routing around
    # the assumption that all classes in the MRO implement the transformation
    # management logic.

class PreservedBySwap(CrossKernel):

    def __new__(cls, *args, **kw):
        if cls is __class__:
            raise TypeError(f'cannot instantiate {__class__.__name__} directly')
        return super().__new__(cls, *args, **kw)

    def _swap(self):
        return super()._swap()._clone(self.__class__)

    # TODO when I implement transformations with methods, make this not an
    # instance of CrossKernel.



# lsqfitgp/_kernels/_wendland.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

from jax import numpy as jnp

from .. import _jaxext
from .._Kernel import stationarykernel, isotropickernel

def _wendland_derivable(k=0, **_):
    return k

def _wendland_maxdim(k=0, alpha=1):
    with _jaxext.skipifabstract():
        return int(jnp.floor(2 * alpha - 1))

@isotropickernel(input='posabs', derivable=_wendland_derivable, maxdim=_wendland_maxdim)
def Wendland(r, k=0, alpha=1):
    """
    Wendland kernel.
    
    .. math::
        k(r) &= \\frac1{B(2k+1,\\nu)}
        \\int_r^\\infty \\mathrm du\\, (u^2 - r^2)^k (1 - u)_+^{\\nu-1}, \\\\
        \\quad k &\\in \\mathbb N,\\ \\nu = k + \\alpha,\\ \\alpha \\ge 1.
    
    An isotropic kernel with finite support. The covariance is nonzero only
    when the distance between the points is less than 1. Parameter :math:`k \\in \\{0,
    1, 2, 3\\}` sets the differentiability, while the maximum dimensionality the
    kernel can be used in is :math:`\\lfloor 2\\alpha-1 \\rfloor`. Default is
    :math:`k = 0` (non derivable), :math:`\\alpha = 1` (can be used only in
    1D).
    
    Reference: Gneiting (2002), Wendland (2004, p. 128), Rasmussen and Williams
    (2006, p. 87), Porcu, Furrer and Nychka (2020, p. 4).
    
    """
        
    # TODO compute the kernel only on the nonzero points.
    
    # TODO find the nonzero points in O(nlogn) instead of O(n^2) by sorting
    # the inputs, and output a sparse matrix => on second thought this should
    # be a general mechanism implemented in GP that gives sparse x and y to
    # the kernel
    
    with _jaxext.skipifabstract():
        D = _wendland_maxdim(k, alpha)
        assert D >= 1, D
    
    if k == 0:
        poly = [
            [1],
        ]
    elif k == 1:
        poly = [
            [1, 1],
            [1],
        ]
    elif k == 2:
        poly = [
            [1/3, 4/3, 1],
            [1, 2],
            [1],
        ]
    elif k == 3:
        poly = [
            [1/15, 3/5, 23/15, 1],
            [2/5, 12/5, 3],
            [1, 3],
            [1],
        ]
    else:
        raise NotImplementedError
    
    nu = k + alpha
    coeffs = jnp.array([jnp.polyval(jnp.array(pj), nu) for pj in poly])
    poly = jnp.polyval(coeffs, r)
    return jnp.where(r < 1, (1 - r) ** (nu + k) * poly, 0)


# adapted from the "GP-circular" example in the PyMC documentation

# TODO maxdim actually makes sense only for isotropic. I need a way to say
# structured/non structured. Maybe all this should just live in the test suite.

# TODO Any stationary kernel supported on [0, pi] would be fine combined with
# the geodesic distance. Use the generic wendland kernel. Options:
# 1) add here the parameters of Wendland
# 2) add a distance option in stationary to use the angular distance, then
#    let the user apply it to wendland => the problem is that the user would
#    need to be careful with positiveness, while wendland checks it for him

@stationarykernel(derivable=1, maxdim=1, input='posabs')
def Circular(delta, tau=4, c=1/2):
    """
    Circular kernel.
    
    .. math:: k(x, y) &= W_c(d_{\\text{geo}}(x, y)), \\\\
        W_c(t) &= \\left(1 + \\tau\\frac tc\\right)
            \\left(1 - \\frac tc\\right)^\\tau_+,
        \\quad c \\in (0, 1/2], \\tau \\ge 4, \\\\
        d_{\\text{geo}}(x, y) &= \\arccos\\cos(2\\pi(x-y)).
    
    It is a stationary periodic kernel with period 1.
    
    Reference: Padonou and Roustant (2016).
    """
    with _jaxext.skipifabstract():
        assert tau >= 4, tau
        assert 0 < c <= 1/2, c
    x = delta % 1
    t = jnp.minimum(x, 1 - x)
    return (1 + tau * t / c) * jnp.maximum(1 - t / c, 0) ** tau



# lsqfitgp/_patch_gvar.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" modifications to the global state of gvar """

import functools

import gvar
from jax import numpy as jnp
from jax.scipy import special as jspecial

gvar_ufuncs = [
    'sin',
    'cos',
    'tan',
    'exp',
    'log',
    'sqrt',
    'fabs',
    'sinh',
    'cosh',
    'tanh',
    'arcsin',
    'arccos',
    'arctan',
    'arctan2',
    'arcsinh',
    'arccosh',
    'arctanh',
    'square',
    'erf',
]

for fname in gvar_ufuncs:
    fgvar = getattr(gvar, fname)
    fjax = getattr(jnp, fname, getattr(jspecial, fname, NotImplemented))
    fboth = functools.singledispatch(fgvar)
    fboth.register(jnp.ndarray, fjax)
    setattr(gvar, fname, fboth)

# reset transformations to support jax arrays
gvar.BufferDict.del_distribution('log')
gvar.BufferDict.del_distribution('sqrt')
gvar.BufferDict.del_distribution('erfinv')
gvar.BufferDict.add_distribution('log', gvar.exp)
gvar.BufferDict.add_distribution('sqrt', gvar.square)
gvar.BufferDict.add_distribution('erfinv', gvar.erf)



# lsqfitgp/_kernels/_matern.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

from jax import numpy as jnp

from .. import _jaxext
from .. import _special
from .._Kernel import isotropickernel

def _maternp_derivable(p=None):
    return p

@isotropickernel(derivable=_maternp_derivable)
def Maternp(r2, p=None):
    """
    Matérn kernel of half-integer order. 
    
    .. math::
        k(r) &= \\frac {2^{1-\\nu}} {\\Gamma(\\nu)} x^\\nu K_\\nu(x) = \\\\
        &= \\exp(-x) \\frac{p!}{(2p)!}
        \\sum_{i=0}^p \\frac{(p+i)!}{i!(p-i)!} (2x)^{p-i} \\\\
        \\nu &= p + 1/2,
        p \\in \\mathbb N,
        x = \\sqrt{2\\nu} r
    
    The degree of derivability is :math:`p`.

    Reference: Rasmussen and Williams (2006, p. 85).
    """
    with _jaxext.skipifabstract():
        assert int(p) == p and p >= 0, p
    r2 = (2 * p + 1) * r2
    return _special.kvmodx2_hi(r2 + 1e-30, p)
    # TODO see if I can remove the 1e-30 improving kvmodx2_hi_jvp

def _matern_derivable(nu=None):
    with _jaxext.skipifabstract():
        return int(max(0, jnp.ceil(nu) - 1))

@isotropickernel(derivable=_matern_derivable)
def Matern(r2, nu=None):
    """
    Matérn kernel of real order. 
    
    .. math::
        k(r) = \\frac {2^{1-\\nu}} {\\Gamma(\\nu)} x^\\nu K_\\nu(x),
        \\quad \\nu \\ge 0,
        \\quad x = \\sqrt{2\\nu} r
    
    The process is :math:`\\lceil\\nu\\rceil-1` times derivable: so for
    :math:`0 \\le \\nu \\le 1` it is not derivable, for :math:`1 < \\nu \\le 2`
    it is derivable but has not a second derivative, etc. The highest
    derivative is (Lipschitz) continuous iff :math:`\\nu\\bmod 1 \\ge 1/2`.

    Reference: Rasmussen and Williams (2006, p. 84).
    """
    with _jaxext.skipifabstract():
        assert 0 <= nu < jnp.inf, nu
    r2 = 2 * jnp.where(nu, nu, 1) * r2  # for v = 0 the correct limit is white
                                        # noise, so I avoid doing r2 * 0
    return _special.kvmodx2(nu, r2)
    
    # TODO broken for high nu. However the convergence to ExpQuad is extremely
    # slow. Tentative temporary patch:
    # - for large x, when x^v=inf, use https://dlmf.nist.gov/10.25.E3
    # - for small x, when Kv(x)=inf, return 1
    # - for very large v, use expquad even if it's not good enough
    
    # The GSL has log K_nu
    # https://www.gnu.org/software/gsl/doc/html/specfunc.html#irregular-modified-bessel-functions-fractional-order
    
# def _bessel_scale(nu):
#     lnu = numpy.floor(nu)
#     rnu = numpy.ceil(nu)
#     zl, = special.jn_zeros(lnu, 1)
#     if lnu == rnu:
#         return zl
#     else:
#         zr, = special.jn_zeros(rnu, 1)
#         return zl + (nu - lnu) * (zr - zl) / (rnu - lnu)

def _bessel_derivable(nu=0):
    with _jaxext.skipifabstract():
        return int(nu // 2)

# TODO looking at the plot in the reference, it seems derivable also for nu = 0.
# what's up? investigate numerically by overwriting the derivability. rasmussen
# does not say anything about it. Problem: my custom derivatives may not work
# properly in this case.

def _bessel_maxdim(nu=0):
    with _jaxext.skipifabstract():
        return 2 * int(jnp.floor(nu) + 1)

@isotropickernel(derivable=_bessel_derivable, maxdim=_bessel_maxdim)
def Bessel(r2, nu=0):
    """
    Bessel kernel.
    
    .. math:: k(r) = \\Gamma(\\nu + 1) 2^\\nu (sr)^{-\\nu} J_{\\nu}(sr),
        \\quad s = 2 + \\nu / 2,
        \\quad \\nu \\ge 0,
    
    where :math:`s` is a crude estimate of the half width at half maximum of
    :math:`J_\\nu`. Can be used in up to :math:`2(\\lfloor\\nu\\rfloor + 1)`
    dimensions and derived up to :math:`\\lfloor\\nu/2\\rfloor` times.
    
    Reference: Rasmussen and Williams (2006, p. 89).
    """
    with _jaxext.skipifabstract():
        assert 0 <= nu < jnp.inf, nu
    r2 = r2 * (2 + nu / 2) ** 2
    return _special.gamma(nu + 1) * _special.jvmodx2(nu, r2)

    # nu >= (D-2)/2
    # 2 nu >= D - 2
    # 2 nu + 2 >= D
    # D <= 2 (nu + 1)



# lsqfitgp/_linalg/_decomp.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

Copy-pasted from the notes:

2023-02-14
==========

My current decomposition system is a mess. I can't take reverse gradients.
I can't straightforwardly implement optimized algorithms that compute together
likelihood, gradient, and fisher. Jax patterns break down unpredictably. I
have to redesign it from scratch.

Guidelines and requirements:

  - Sufficient modularity to implement composite decompositions (Woodbury,
    Block)

  - Does not mess up jax in any way

  - Caches decompositions

  - Favors optimizing together the likelihood and its derivatives

Operations (in the following I indicate with lowercase inputs which are
typically vectors or tall matrices, and uppercase inputs which are typically
large matrices, since optimization requires taking it into account):

    pinv_bilinear(A, r) => A'K⁺r (for the posterior mean)
    pinv_bilinear_robj(A, r) same but r can be gvars
    ginv_quad(A) => A'K⁻A (for the posterior covariance)
        I want the pseudoinverse for the mean because the data may not be
        in the span and I want to project it orthogonally, while for the
        covariance I expect A and K to come from a pos def matrix so they are
        coherent
    ginv_diagquad(A) => diag(A'K⁻A) (for the posterior variance)
    minus_log_normal_density(
        r: 1d array,      # the residuals (data - prior mean)
        dr_vjp: callable, # x -> x_i ∂r_i/∂p_j,   gradrev and fishvec
        dK_vjp: callable, # x -> x_ij ∂K_ij/∂p_k, gradrev and fishvec
        dr_jvp: callable, # x -> ∂r_i/∂r_j x_j,  fishvec
        dK_jvp: callable, # x -> ∂K_ij/∂p_k x_k, fishvec
        dr: 2d array,     # ∂r_i/∂p_j,  gradfwd and fisher
        dK: 3d array,     # ∂K_ij/∂p_k, gradfwd and fisher
        vec: 1d array,    # input vector of fishvec, same size as params
        value: bool,
        gradrev: bool,
        gradfwd: bool,
        fisher: bool,
        fishvec: bool,
    )
        This computes on request
            value: 1/2 tr(KK⁺) log 2π
                 + 1/2 tr(I-KK⁺) log 2π
                 + 1/2 log pdet K
                 + 1/2 tr(I-KK⁺) log ε
                 + 1/2 r'(K⁺+(I-KK⁺)/ε)r
            gradrev,
            gradfwd: 1/2 tr(K⁺dK)
                    + r'(K⁺+(I-KK⁺)/ε) dr
                    - 1/2 r'(K⁺+2(I-KK⁺)/ε)dKK⁺r
            fisher: 1/2 tr(K⁺dK(K⁺+2(I-KK⁺)/ε)d'K)
                  - 2 tr(K⁺dK(I-KK⁺)d'KK⁺)
                  + dr'(K⁺+(I-KK⁺)/ε)d'r
            fishvec: fisher matrix times vec
        There should be options for omitting the pieces with ε. I also need a
        way to make densities with different values of ε comparable with each
        other (may not be possible, if it is, it probably requires a history of
        ranks and ε). gradfwd/rev form K⁺ explicitly to compute tr(K⁺dK) for
        efficiency.
    correlate(x)
        Zx where K = ZZ'.
    back_correlate(X):
        Z'X, this is used by Sandwich and Woodbury.

Since I also want to compute the Student density, I could split
minus_log_normal_density's return value into logdet and quad. The gradient
splits nicely between the two terms, but I have to redo the calculation of the
Fisher matrix for the Student distribution. Alternatively, I could use the the
Normal Fisher. => See Lange et al. (1989, app. B). => I think I can split the
gradient and Fisher matrix too.

2023-03-07
==========

To compute a Fisher-vector product when there are many parameters, do

    tr(K+ dK K+ dK) v =
    = K_vjp(K+ K_jvp(v) K+)

"""

# TODO to automatize this further, I could take in a function that generates K
# (or its pieces) and the arguments to the function. But how would this play
# together with passing decomposition objects as pieces?

# TODO split this file by class

import abc
import functools

import numpy
import jax
from jax import numpy as jnp
from jax.scipy import linalg as jlinalg
from jax import lax

from .. import _jaxext
from . import _pytree

class Decomposition(_pytree.AutoPyTree, abc.ABC):
    """
    Abstract base class for decompositions of positive semidefinite matrices.
    """

    @abc.abstractmethod
    def __init__(self, *args, **kw):
        """ Decompose the input matrix """
        pass

    @abc.abstractmethod
    def matrix(self):
        """ The input matrix """
        pass

    @abc.abstractmethod
    def ginv_linear(self, X):
        """ Compute K⁻X """
        pass

    @abc.abstractmethod
    def pinv_bilinear(self, A, r):
        """Compute A'K⁺r."""
        pass

    @abc.abstractmethod
    def pinv_bilinear_robj(self, A, r):
        """Compute A'K⁺r, where r can be an array of objects."""
        pass

    @abc.abstractmethod
    def ginv_quad(self, A):
        """Compute A'K⁻A."""
        pass

    @abc.abstractmethod
    def ginv_diagquad(self, A):
        """Compute diag(A'K⁻A)."""
        pass

    @abc.abstractmethod
    def correlate(self, x):
        """ Compute Zx where K = ZZ' """
        pass

    @abc.abstractmethod
    def back_correlate(self, X):
        """ Compute Z'X """
        pass

    @abc.abstractmethod
    def pinv_correlate(self, x):
        """ Compute Z⁺x """
        pass

    @abc.abstractmethod
    def minus_log_normal_density(self,
        r,
        *,
        dr_vjp=None,
        dK_vjp=None,
        dr_jvp=None,
        dK_jvp=None,
        dr=None,
        dK=None,
        value=False,
        gradrev=False,
        gradfwd=False,
        fisher=False,
        fishvec=False,
        ):
        """
        Compute minus log a Normal density and its derivatives, with covariance
        matrix K.

        If an input derivative is not specified, it is assumed to be zero.

        Parameters
        ----------
        r: 1d array
            The residuals (value - mean)
        dr_vjp: callable
            x -> x_i ∂r_i/∂p_j, for gradrev and fishvec
        dK_vjp: callable
            x -> x_ij ∂K_ij/∂p_k, for gradrev and fishvec
        dr_jvp_vec: 1d array
            ∂r_i/∂r_j vec_j, for fishvec
        dK_jvp_vec: 2d array
            ∂K_ij/∂p_k vec_k, for fishvec
        dr: 2d array
            ∂r_i/∂p_j for gradfwd and fisher
        dK: 3d array
            ∂K_ij/∂p_k, for gradfwd and fisher
        value: bool
        gradrev: bool
        gradfwd: bool
        fisher: bool
        fishvec: bool
            These parameters indicate which of the return values to compute.
            Default all False.

        Returns
        -------
        value: 1/2 tr(KK⁺) log 2π
             + 1/2 tr(I-KK⁺) log 2π
             + 1/2 log pdet K
             + 1/2 tr(I-KK⁺) log ε
             + 1/2 r'(K⁺+(I-KK⁺)/ε)r
        gradrev,
        gradfwd: 1/2 tr(K⁺dK)
                + r'(K⁺+(I-KK⁺)/ε) dr
                - 1/2 r'(K⁺+2(I-KK⁺)/ε)dKK⁺r
        fisher: 1/2 tr(K⁺dK(K⁺+2(I-KK⁺)/ε)d'K)
              - 2 tr(K⁺dK(I-KK⁺)d'KK⁺)
              + dr'(K⁺+(I-KK⁺)/ε)d'r
        fishvec: fisher matrix @ vec
        """
        pass

    def _parseeps(self, K, epsrel, epsabs, maxeigv=None):
        """ Determine eps from input arguments """
        machine_eps = jnp.finfo(_jaxext.float_type(K)).eps
        if epsrel == 'auto':
            epsrel = len(K) * machine_eps
        if epsabs == 'auto':
            epsabs = machine_eps
        if maxeigv is None:
            maxeigv = eigval_bound(K)
        self._eps = epsrel * maxeigv + epsabs
        return self._eps

    @property
    def eps(self):
        """
        The threshold below which eigenvalues are too small to be determined.
        """
        return self._eps

    @property
    @abc.abstractmethod
    def n(self):
        """ Number of rows/columns of the matrix """
        pass

    @property
    @abc.abstractmethod
    def m(self):
        """ Number of columns of Z """
        pass

    def ginv(self):
        """ Compute K⁻ """
        return self.ginv_quad(jnp.eye(self.n))

def solve_triangular_python(a, b, *, lower=False):
    """
    Pure python implementation of scipy.linalg.solve_triangular for when
    a or b are object arrays.
    """
    # TODO maybe commit this to gvar.linalg
    a = numpy.asarray(a)
    x = numpy.copy(b)

    vec = x.ndim < 2
    if vec:
        x = x[:, None]

    n = a.shape[-1]
    assert x.shape[-2] == n

    if not lower:
        a = a[..., ::-1, ::-1]
        x = x[..., ::-1, :]

    x[..., 0, :] /= a[..., 0, 0, None]
    for i in range(1, n):
        x[..., i:, :] -= x[..., None, i - 1, :] * a[..., i:, i - 1, None]
        x[..., i, :] /= a[..., i, i, None]

    if not lower:
        x = x[..., ::-1, :]

    if vec:
        x = numpy.squeeze(x, -1)
    return x

def solve_triangular_batched(a, b, *, lower=False):
    """ Version of jax.scipy.linalg.solve_triangular that batches matmul-like """
    a = jnp.asarray(a)
    b = jnp.asarray(b)
    vec = b.ndim < 2
    if vec:
        b = b[:, None]

    batch_shape = jnp.broadcast_shapes(a.shape[:-2], b.shape[:-2])
    a_shape = batch_shape + a.shape[-2:]
    b_shape = batch_shape + b.shape[-2:]
    result = lax.linalg.triangular_solve(
        jnp.broadcast_to(a, a_shape), jnp.broadcast_to(b, b_shape),
        left_side=True, lower=lower,
    )
    assert result.shape == b_shape

    if vec:
        result = result.squeeze(-1)
    return result

def solve_batched(a, b, **kw):
    """ Version of jax.scipy.linalg.solve that batches matmul-like """
    a = jnp.asarray(a)
    b = jnp.asarray(b)
    vec = b.ndim < 2
    if vec:
        b = b[:, None]

    @functools.partial(jnp.vectorize, signature='(i,j),(j,k)->(i,k)')
    def solve_batched(a, b):
        return jlinalg.solve(a, b, **kw)
    result = solve_batched(a, b)

    if vec:
        result = result.squeeze(-1)
    return result

def eigval_bound(K):
    """
    Upper bound on the largest magnitude eigenvalue of the matrix, from
    Gershgorin's theorem.
    """
    return jnp.max(jnp.sum(jnp.abs(K), axis=1))

def diag_scale_pow2(K):
    """
    Compute a vector s of powers of 2 such that diag(K / outer(s, s)) ~ 1.
    """
    d = jnp.diag(K)
    return jnp.where(d, jnp.exp2(jnp.rint(0.5 * jnp.log2(d))), 1)

def transpose(x):
    """ swap the last two axes of array x, corresponds to matrix tranposition
    with the broadcasting convention of matmul """
    if x.ndim < 2:
        return x
    elif isinstance(x, jnp.ndarray):
        return jnp.swapaxes(x, -2, -1)
    else:
        # need to support numpy because this function is used with gvars
        return numpy.swapaxes(x, -2, -1)

class Chol(Decomposition):
    """Cholesky decomposition. The matrix is regularized adding a small multiple
    of the identity."""

    def __init__(self, K, *, epsrel='auto', epsabs=0):
        # K <- K + Iε
        # K = LL'
        self._K = K
        s = diag_scale_pow2(K)
        K = K / s / s[:, None]
        eps = self._parseeps(K, epsrel, epsabs)
        K = K.at[jnp.diag_indices_from(K)].add(eps)
        L = jlinalg.cholesky(K, lower=True)
        with _jaxext.skipifabstract():
            if not jnp.all(jnp.isfinite(L)):
                # TODO check that jax fills with nan after failed row, detect
                # and report minor index like scipy
                raise numpy.linalg.LinAlgError('cholesky decomposition not finite, probably matrix not pos def numerically')
        self._L = L * s[:, None]
        self._eps = eps * jnp.min(s * s)

    def matrix(self):
        return self._K

    def ginv_linear(self, X):
        # = K⁻¹X
        # K⁻¹ = L'⁻¹L⁻¹
        # K⁻¹X = L'⁻¹(L⁻¹X)
        invLX = jlinalg.solve_triangular(self._L, X, lower=True)
        return jlinalg.solve_triangular(self._L.T, invLX, lower=False)

    def pinv_bilinear(self, A, r):
        # = A'K⁻¹r = A'L'⁻¹L⁻¹r = (L⁻¹A)'(L⁻¹r)
        invLr = jlinalg.solve_triangular(self._L, r, lower=True)
        invLA = jlinalg.solve_triangular(self._L, A, lower=True)
        return invLA.T @ invLr

    def pinv_bilinear_robj(self, A, r):
        # = A'K⁻¹r
        invLr = solve_triangular_python(self._L, r, lower=True)
        invLA = jlinalg.solve_triangular(self._L, A, lower=True)
        return numpy.asarray(invLA).T @ invLr

    def ginv_quad(self, A):
        # = A'K⁻¹A = A'K⁻¹A = A'L'⁻¹L⁻¹A = (L⁻¹A)'(L⁻¹A)
        invLA = jlinalg.solve_triangular(self._L, A, lower=True)
        return invLA.T @ invLA

    def ginv_diagquad(self, A):
        # = diag(A'K⁻¹A)
        # X = L⁻¹A
        # diag(A'K⁻¹A)_i = diag(X'X)_i = ∑_j X'_ij X_ji = ∑_j X_ji X_ji
        invLA = jlinalg.solve_triangular(self._L, A, lower=True)
        return jnp.einsum('ji,ji->i', invLA, invLA)

    def correlate(self, x):
        # = Lx
        return self._L @ x

    def back_correlate(self, X):
        # = L'X
        return self._L.T @ X

    def pinv_correlate(self, x):
        # = L⁻¹x
        return jlinalg.solve_triangular(self._L, x, lower=True)

    def minus_log_normal_density(self,
        r,               # 1d array, the residuals (data - prior mean)
        *,
        dr_vjp=None,     # callable, x -> x_i ∂r_i/∂p_j,   gradrev and fishvec
        dK_vjp=None,     # callable, x -> x_ij ∂K_ij/∂p_k, gradrev and fishvec
        dr_jvp_vec=None, # 1d array, ∂r_i/∂r_j v_j,  fishvec
        dK_jvp_vec=None, # 2d array, ∂K_ij/∂p_k v_k, fishvec
        dr=None,         # 2d array, ∂r_i/∂p_j,  gradfwd and fisher
        dK=None,         # 3d array, ∂K_ij/∂p_k, gradfwd and fisher
        value=False,
        gradrev=False,
        gradfwd=False,
        fisher=False,
        fishvec=False,
    ):

        L = self._L

        out = {}

        # compute shared factors
        grad = (
            (gradrev and (dK_vjp is not None or dr_vjp is not None))
            or (gradfwd and (dK is not None or dr is not None))
        )
        if value or grad:
            invLr = jlinalg.solve_triangular(L, r, lower=True)
        if grad:
            invKr = jlinalg.solve_triangular(L.T, invLr, lower=False)
        if (gradrev and dK_vjp is not None) or (gradfwd and dK is not None):
            invL = jlinalg.solve_triangular(L, jnp.eye(len(L)), lower=True)
            invK = invL.T @ invL

        if value:
            # = 1/2 n log 2π
            #   + 1/2 log det K
            #   + 1/2 r'K⁻¹r
            # K = LL'
            # K⁻¹ = L'⁻¹L⁻¹
            # det K = (det L)² =
            #       = (∏_i L_ii)²
            # r'K⁻¹r = r'L'⁻¹L⁻¹r =
            #        = (L⁻¹r)'(L⁻¹r)
            out['value'] = 1/2 * (
                len(L) * jnp.log(2 * jnp.pi) +
                2 * jnp.sum(jnp.log(jnp.diag(L))) +
                invLr @ invLr
            )
        else:
            out['value'] = None

        if gradrev:
            # = 1/2 tr(K⁻¹dK)
            #   + r'K⁻¹dr
            #   - 1/2 r'K⁻¹dKK⁻¹r
            # tr(K⁻¹dK) = K⁻¹_ij dK_ji =
            #           = K⁻¹_ij dK_ij =
            #           = dK_vjp(K⁻¹)
            # r'K⁻¹dr = r_i K⁻¹_ij dr_j =
            #         = (K⁻¹r)_j dr_j =
            #         = dr_vjp(K⁻¹r)
            # r'K⁻¹dKK⁻¹r = r_i K⁻¹_ij dK_jl K⁻¹_lm r_m =
            #             = (K⁻¹r)_j dK_jl (K⁻¹r)_l =
            #             = dK_vjp((K⁻¹r) ⊗ (K⁻¹r))
            out['gradrev'] = 0
            if dK_vjp is not None:
                tr_invK_dK = dK_vjp(invK)
                r_invK_dK_invK_r = dK_vjp(jnp.outer(invKr, invKr))
                out['gradrev'] += 1/2 * (tr_invK_dK - r_invK_dK_invK_r)
            if dr_vjp is not None:
                r_invK_dr = dr_vjp(invKr)
                out['gradrev'] += r_invK_dr
        else:
            out['gradrev'] = None

        if gradfwd:
            # = 1/2 tr(K⁻¹dK)
            #   + r'K⁻¹dr
            #   - 1/2 r'K⁻¹dKK⁻¹r
            # tr(K⁻¹dK)_k = K⁻¹_ij dK_ijk
            # r'K⁻¹dr = (K⁻¹r)'dr
            # (r'K⁻¹dKK⁻¹r)_k = r_i K⁻¹_ij dK_jlk K⁻¹_lm r_m =
            #                 = (K⁻¹r)_j dK_jlk (K⁻¹r)_l
            out['gradfwd'] = 0
            if dK is not None:
                tr_invK_dK = jnp.einsum('ij,ijk->k', invK, dK)
                r_invK_dK_invK_r = jnp.einsum('i,ijk,j->k', invKr, dK, invKr)
                out['gradfwd'] += 1/2 * (tr_invK_dK - r_invK_dK_invK_r)
            if dr is not None:
                r_invK_dr = invKr @ dr
                out['gradfwd'] += r_invK_dr
        else:
            out['gradfwd'] = None

        if fisher:
            # = 1/2 tr(K⁻¹dKK⁻¹d'K)
            #   + dr'K⁻¹d'r
            # tr(K⁻¹dKK⁻¹d'K)_ij = tr(L'⁻¹L⁻¹dKL'⁻¹L⁻¹d'K)_ij =
            #                    = tr(L⁻¹dKL'⁻¹L⁻¹d'KL'⁻¹)_ij =
            #                    = (L⁻¹dKL'⁻¹)_kli (L⁻¹dKL'⁻¹)_klj
            # (L⁻¹dKL'⁻¹)_ijk = L⁻¹_il dK_lmk L'⁻¹_mj =
            #                 = L⁻¹_il L⁻¹_jm dK_lmk
            # (dr'K⁻¹d'r)_kq = dr'_k L'⁻¹L⁻¹dr_q =
            #                = (L⁻¹dr_k)_i (L⁻¹dr_q)_i
            out['fisher'] = 0
            if dK is not None:
                invL_dK = solve_triangular_batched(L,
                    jnp.moveaxis(dK, 2, 0),
                    lower=True) # kim: L⁻¹_il dK_lmk
                invL_dK_invL = solve_triangular_batched(L,
                    jnp.swapaxes(invL_dK, 1, 2),
                    lower=True) # kji: L⁻¹_jm (L⁻¹_il dK_lmk)
                tr_invK_dK_invK_dK = jnp.einsum('kij,qij->kq', invL_dK_invL, invL_dK_invL)
                out['fisher'] += 1/2 * tr_invK_dK_invK_dK
            if dr is not None:
                invLdr = jlinalg.solve_triangular(L, dr, lower=True)
                dr_invK_dr = invLdr.T @ invLdr
                out['fisher'] += dr_invK_dr
        else:
            out['fisher'] = None

        if fishvec:
            # = 1/2 tr(K⁻¹dKK⁻¹d'K) v
            #   + dr'K⁻¹d'r v
            # tr(K⁻¹dKK⁻¹d'K) v = K_vjp(K⁻¹K_jvp(v)K⁻¹) =
            #                   = K_vjp(L'⁻¹L⁻¹ K_jvp(v) L'⁻¹L⁻¹)
            # dr'K⁻¹d'r v = dr'K⁻¹dr_jvp(v) =
            #             = dr_vjp(K⁻¹dr_jvp(v)) =
            #             = dr_vjp(L'⁻¹L⁻¹ dr_jvp(v))
            out['fishvec'] = 0
            if not (dK_jvp_vec is None and dK_vjp is None):
                invL_dKv = jlinalg.solve_triangular(L, dK_jvp_vec, lower=True)
                invK_dKv = jlinalg.solve_triangular(L.T, invL_dKv, lower=False)
                invL_dKv_invK = jlinalg.solve_triangular(L, invK_dKv.T, lower=True)
                invK_dKv_invK = jlinalg.solve_triangular(L.T, invL_dKv_invK, lower=False)
                tr_invK_dK_invK_dK_v = dK_vjp(invK_dKv_invK)
                out['fishvec'] += 1/2 * tr_invK_dK_invK_dK_v
            if not (dr_jvp_vec is None and dr_vjp is None):
                invL_drv = jlinalg.solve_triangular(L, dr_jvp_vec, lower=True)
                invK_drv = jlinalg.solve_triangular(L.T, invL_drv, lower=False)
                dr_invK_drv_v = dr_vjp(invK_drv)
                out['fishvec'] += dr_invK_drv_v
        else:
            out['fishvec'] = None

        return tuple(out.values())

    @classmethod
    def make_derivs(cls,
        K_fun, r_fun, primal,
        *,
        args=(),
        kw={},
        vec=None,
        value=False,
        gradrev=False,
        gradfwd=False,
        fisher=False,
        fishvec=False,
    ):
        """
        Prepares arguments for `minus_log_normal_density`.

        Parameters
        ----------
        K_fun, r_fun : callable
            Functions with signature ``f(primal, *args, **kw)`` that produce the
            `K` init argument and the `r` `minus_log_normal_density` argument.
        primal : 1d array
            The first argument to `K_fun` and `r_fun`.
        args : tuple
            Additional positional arguments to `K_fun` and `r_fun`.
        kw : dict
            Keyword arguments to `K_fun` and `r_fun`.
        vec : 1d array
            A tangent vector to compute the jacobian-vector products.
        value, gradrev, gradfwd, fisher, fishvec : bool
            Arguments to `minus_log_normal_density`, used to determine which
            derivatives are needed.

        Returns
        -------
        K : 2d array
            Output of `K_fun`.
        r : 1d array
            Output of `r_fun`.
        out : dict
            Dictionary with derivative arguments to `minus_log_normal_density`.
        """

        partial = lambda f: lambda x: f(x, *args, **kw)
        K_fun = partial(K_fun)
        r_fun = partial(r_fun)

        out = {}

        if gradrev or fishvec:
            K, dK_vjp = jax.vjp(K_fun, primal)
            r, dr_vjp = jax.vjp(r_fun, primal)
            out['dK_vjp'] = lambda x: dK_vjp(x)[0]
            out['dr_vjp'] = lambda x: dr_vjp(x)[0]
        else:
            K = K_fun(primal)
            r = r_fun(primal)
        if fishvec:
            _, out['dK_jvp_vec'] = jax.jvp(K_fun, (primal,), (vec,))
            _, out['dr_jvp_vec'] = jax.jvp(r_fun, (primal,), (vec,))
        if gradfwd or fisher:
            out['dK'] = jax.jacfwd(K_fun)(primal)
            out['dr'] = jax.jacfwd(r_fun)(primal)

        return K, r, out

    @property
    def n(self):
        return len(self._L)

    m = n
